"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, Pattern Recognition Lab, 
         Friedrich-Alexander University Erlangen-Nuremberg 
         marc.aubreville@fau.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images
        Bildverarbeitung fuer die Medizin 2018, Springer Verlag, Berlin-Heidelberg

"""

__all__ = ['database', 'slide', 'dicomimage', 'annotations', 'nifty']

from .slide import os_fileformats
from .nifty import nii_fileformats
from .dicomimage import dic_fileformats

fileformats = os_fileformats + nii_fileformats + dic_fileformats




"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Slide wrapper for handling multiple slide formats (including multiprocessing to have slide processing in separate thread)


"""
import multiprocessing
import queue
import time
import numpy as np
import os
from . import dicomimage
from . import nifty
from . import cellvizio
import openslide
from openslide.lowlevel import OpenSlideError, OpenSlideUnsupportedFormatError
os_fileformats = ['*.svs','*.tif', '*.png', '*.bif', '*.svslide', '*.mrxs' ,'*.scn' ,'*.vms' ,'*.vmu', '*.ndpi', '*.tiff', '*.bmp']
from PIL import Image
import PIL
PIL.Image.MAX_IMAGE_PIXELS = 933120000

# This is a 3D version of openslide.ImageSlide, supporting z stacks

class ImageSlide3D(openslide.ImageSlide):

    def __init__(self, file):
        """Open an image file.

        file can be a filename or a PIL.Image."""
        openslide.ImageSlide.__init__(self, file)

        try:
            self.numberOfFrames = self._image.n_frames 
        except:
            self.numberOfFrames = 1
        


    def read_region(self, location, level, size, zStack):
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the level 0
                  reference frame.
        level:    the level number.
        size:     (width, height) tuple giving the region size."""

        if level != 0:
            raise OpenSlideError("Invalid level")
        if ['fail' for s in size if s < 0]:
            raise OpenSlideError("Size %s must be non-negative" % (size,))
        # Any corner of the requested region may be outside the bounds of
        # the image.  Create a transparent tile of the correct size and
        # paste the valid part of the region into the correct location.
        image_topleft = [max(0, min(l, limit - 1))
                    for l, limit in zip(location, self._image.size)]
        image_bottomright = [max(0, min(l + s - 1, limit - 1))
                    for l, s, limit in zip(location, size, self._image.size)]
        tile = Image.new("RGBA", size, (0,) * 4)

        if not ['fail' for tl, br in zip(image_topleft, image_bottomright)
                if br - tl < 0]:  # "< 0" not a typo
            # Crop size is greater than zero in both dimensions.
            # PIL thinks the bottom right is the first *excluded* pixel
            self._image.seek(zStack)
            crop = self._image.crop(image_topleft +
                    [d + 1 for d in image_bottomright])
            tile_offset = tuple(il - l for il, l in
                    zip(image_topleft, location))
            tile.paste(crop, tile_offset)
        return tile    

class RotatableOpenSlide(object):

    def __new__(cls, filename, rotate):
        if cls is RotatableOpenSlide:
            bname, ext = os.path.splitext(filename)
            if ext.upper() == '.DCM': return type("RotatableOpenSlide", (RotatableOpenSlide, dicomimage.ReadableDicomDataset,openslide.ImageSlide), {})(filename, rotate)
            if (ext.upper() in ['.NII','.GZ']): 
                return type("ReadableNIBDataset", (RotatableOpenSlide,nifty.ReadableNIBDataset, openslide.ImageSlide ), {})(filename, rotate)
            if ext.upper() == '.MKT': return type("ReadableCellVizioMKTDataset", (RotatableOpenSlide, cellvizio.ReadableCellVizioMKTDataset, openslide.ImageSlide), {})(filename,rotate)
            if ext.upper() == '.TIF': return type("RotatableOpenSlide", (RotatableOpenSlide, ImageSlide3D,openslide.ImageSlide), {})(filename, rotate)
            try:
                slideobj = type("OpenSlide", (RotatableOpenSlide,openslide.OpenSlide), {})(filename, rotate)
                slideobj.isOpenSlide = True
                return slideobj
            except:
                slideobj = type("ImageSlide", (RotatableOpenSlide,ImageSlide3D), {})(filename, rotate)
                slideobj.isOpenSlide = False
                return slideobj
        else:
            return object.__new__(cls)

    def __init__(self, filename, rotate=False):
        if ('rotate' in self.__dict__): # speed up - somehow init is called twice. Let's skip that.
            return
        self.rotate=rotate
        self.type=0
        self.numberOfFrames = 1
        self.fps = 1.0
        return super().__init__(filename)



    # Implements 180 degree rotated version of read_region
    def read_region(self, location, level, size, zLevel=0):
        # zlevel is ignored for SVS files
        if (self.rotate):
            location = [int(x-y-(w*self.level_downsamples[level])) for x,y,w in zip(self.dimensions, location, size)]
            if (self.isOpenSlide):
                return super().read_region(location, level, size).rotate(180)
            else:
                return super().read_region(location, level, size, zLevel=zLevel).rotate(180)
        else:
            if (self.isOpenSlide):
                return super().read_region(location, level, size)
            else:
                return super().read_region(location, level, size, zLevel)

    def transformCoordinates(self, location, level=0, size=None, inverse=False):
        if (self.rotate):
            retarr = np.copy(location)
            retarr[:,0] = self.dimensions[0]-retarr[:,0]
            retarr[:,1] = self.dimensions[1]-retarr[:,1]
            return retarr
        else:
            return location

    def get_label_image(self, target_size_x):
        if (self.isOpenSlide):
            if 'label' in super().associated_images:
                img = super().associated_images['label']
                img = img.resize((target_size_x, int(img.size[1]/img.size[0]*target_size_x)))
                return np.array(img)
        
        return None

    def slide_center(self):
        return [int(x/2) for x in self.dimensions]
    
    def read_centerregion(self, location, level, size, center=None, zLevel=0):
        center = self.slide_center() if center is None else center
        return self.read_region([int(x-s*self.level_downsamples[level]/2-d) for x,d,s in zip(center,location, size)], level, size, zLevel)
    

class SlideReader(multiprocessing.Process):
    def __init__(self):
        multiprocessing.Process.__init__(self)
        self.sl = None
        self.slidename = None
        self.slide = None
        self.daemon=True
        self.queue = multiprocessing.Queue(50)
        self.outputQueue = multiprocessing.Queue()

    def run(self):
        img=None
        lastReq = [(-1,-1),-1,(512,512)]
        while (True):
            (slidename, location, level, size, id, rotated, zlevel) = self.queue.get()

            try:
                while(True):
                    (slidename, location, level, size, id, rotated, zlevel) = self.queue.get(True,0.01)
            except queue.Empty:
                pass


            if (slidename==-1):
                print('Exiting SlideReader thread')
                return

            if (slidename!=self.slidename):
                self.slide = RotatableOpenSlide(slidename, rotate=rotated)
                self.slidename = slidename
                newSlide=True
            else:
                newSlide=False

            self.slide.rotate = rotated

            if not all([a==b for a,b in zip([location,level,size,zlevel,rotated],lastReq)]) or newSlide:
                img = self.slide.read_region(location, level, size, zLevel=zlevel)
                lastReq = [location, level, size,zlevel,rotated]

            self.outputQueue.put((np.array(img),id))




"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Support for (SQLITE) databases 


"""

import sqlite3
import os
import time
import numpy as np
import random
import uuid
import sys
from typing import Dict
from .annotations import *



def hex_to_rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

def generate_uuid():
    return str(uuid.uuid4())

def rgb_to_hex(col):
    """
        Assigns one of sliderunners colors to a class
    """
    return '#{:02x}{:02x}{:02x}'.format( *col[0:3] )

def random_color():
    return rgb_to_hex([random.randrange(30,255),random.randrange(30,255),random.randrange(30,255)])

class DatabaseField(object):
    def __init__(self, keyStr:str, typeStr:str, isNull:int=0, isUnique:bool=False, isAutoincrement:bool=False, defaultValue:str='', primaryKey:int=0):
        self.key = keyStr
        self.type = typeStr
        self.isNull = isNull
        self.isUnique = isUnique
        self.isAutoincrement = isAutoincrement
        self.defaultValue = defaultValue
        self.isPrimaryKey = primaryKey

    def creationString(self):
        return ("`"+self.key+"` "+self.type+
                ((" DEFAULT %s " % self.defaultValue) if not(self.defaultValue == "") else "") + 
                (" PRIMARY KEY" if self.isPrimaryKey else "")+
                (" AUTOINCREMENT" if self.isAutoincrement else "")+
                (" UNIQUE" if self.isUnique else "") )


def isActiveClass(label, activeClasses):
        if (label) is None:
            return True
        if (label<len(activeClasses)):
            return activeClasses[label]
        else:
            print('Warning: Assigned label is: ',label,' while class list is: ',activeClasses)
            return 0

class DatabaseTable(object):
    def __init__(self, name:str):
        self.entries = dict()
        self.name = name

    def add(self, field:DatabaseField):
        self.entries[field.key] = field
        return self

    def getCreateStatement(self):
        createStatement = "CREATE TABLE `%s` (" % self.name
        cnt=0
        for idx, entry in self.entries.items():
            if cnt>0:
                createStatement += ','
            createStatement += entry.creationString()
            cnt+=1
        createStatement += ");"
        return createStatement

    def checkTableInfo(self, tableInfo):
        if (tableInfo is None) or (len(tableInfo)==0):
            return False
        allKeys=list()
        for entry in tableInfo:
            idx,key,typeStr,isNull,defaultValue,PK = entry
            if (key not in self.entries):
                return False
            if not (self.entries[key].type == typeStr):
                return False
            allKeys += [key]
        for entry in self.entries:
            if entry not in allKeys:
                return False
        # OK, defaultValue, isNull are not important to be checked

        return True


    def addMissingTableInfo(self, tableInfo, tableName):
        returnStr = list()
        if (tableInfo is None) or (len(tableInfo)==0):
            return False
        allKeys=list()
        for entry in tableInfo:
            idx,key,typeStr,isNull,defaultValue,PK = entry
            allKeys += [key]

        for entry in self.entries.keys():
            if entry not in allKeys:
                returnStr.append('ALTER TABLE %s ADD COLUMN %s' % (tableName, self.entries[entry].creationString()))

        return returnStr



class Database(object):
    annotations = Dict[int,annotation]

    minCoords = np.empty((0,3))
    maxCoords = np.empty((0,3))

    def __init__(self):
        self.dbOpened = False
        self.VA = dict()

        self.databaseStructure = dict()
        self.transformer = None
        self.annotations = dict()       
        self.doCommit = True
        self.annotationsSlide = None
        self.exactUser = 0
        self.databaseStructure['Log'] = DatabaseTable('Log').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('dateTime','FLOAT')).add(DatabaseField('labelId','INTEGER'))
        self.databaseStructure['Slides'] = DatabaseTable('Slides').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('filename','TEXT')).add(DatabaseField('width','INTEGER')).add(DatabaseField('EXACTUSER','INTEGER',defaultValue=0)).add(DatabaseField('height','INTEGER')).add(DatabaseField('directory','TEXT')).add(DatabaseField('uuid','TEXT')).add(DatabaseField('exactImageID', 'TEXT'))
        self.databaseStructure['Annotations'] = DatabaseTable('Annotations').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('guid','TEXT')).add(DatabaseField('deleted','INTEGER',defaultValue=0)).add(DatabaseField('slide','INTEGER')).add(DatabaseField('type','INTEGER')).add(DatabaseField('agreedClass','INTEGER')).add(DatabaseField('lastModified','REAL',defaultValue=str(time.time()))).add(DatabaseField('description','TEXT')).add(DatabaseField('clickable', 'INTEGER', defaultValue=1))
        self.databaseStructure['Annotations_coordinates'] = DatabaseTable('Annotations_coordinates').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('coordinateX','INTEGER')).add(DatabaseField('coordinateY','INTEGER')).add(DatabaseField('coordinateZ','INTEGER',defaultValue=0)).add(DatabaseField('slide','INTEGER')).add(DatabaseField('annoId','INTEGER')).add(DatabaseField('orderIdx','INTEGER'))
        self.databaseStructure['Persons'] = DatabaseTable('Persons').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('name','TEXT')).add(DatabaseField('isExactUser','INTEGER', defaultValue=0))
        self.databaseStructure['Classes'] = DatabaseTable('Classes').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('name','TEXT')).add(DatabaseField('clickable','INTEGER', defaultValue=1))
        self.databaseStructure['Annotations_label'] = DatabaseTable('Annotations_label').add(DatabaseField('uid','INTEGER',isAutoincrement=True, primaryKey=1)).add(DatabaseField('exact_id','INTEGER')).add(DatabaseField('person','INTEGER',defaultValue=0)).add(DatabaseField('class','INTEGER')).add(DatabaseField('annoId','INTEGER'))

    def isOpen(self):
        return self.dbOpened

    def appendToMinMaxCoordsList(self, anno: annotation):
        self.minCoords = np.vstack((self.minCoords, np.asarray(anno.minCoordinates().tolist()+[anno.uid])))
        self.maxCoords = np.vstack((self.maxCoords, np.asarray(anno.maxCoordinates().tolist()+[anno.uid])))
    def generateMinMaxCoordsList(self):
        # MinMaxCoords lists shows extreme coordinates from object, to decide if an object shall be shown
        self.minCoords = np.zeros(shape=(len(self.annotations),3))
        self.maxCoords = np.zeros(shape=(len(self.annotations),3))
        keys = self.annotations.keys() 
        for idx,annokey in enumerate(keys):
            annotation = self.annotations[annokey]
            self.minCoords[idx] = np.asarray(annotation.minCoordinates().tolist()+[annokey])
            self.maxCoords[idx] = np.asarray(annotation.maxCoordinates().tolist()+[annokey])

    def getVisibleAnnotations(self, leftUpper:list, rightLower:list) -> Dict[int, annotation]:
        potentiallyVisible =  ( (self.maxCoords[:,0] > leftUpper[0]) & (self.minCoords[:,0] < rightLower[0]) & 
                                (self.maxCoords[:,1] > leftUpper[1]) & (self.minCoords[:,1] < rightLower[1]) )
        ids = self.maxCoords[potentiallyVisible,2]
        return dict(filter(lambda i:i[0] in ids, self.annotations.items()))

    def setClassColor(self, classId, color:str):
        """
            color: valid hex rgb string, e.g. #AABBCC
            classId: Class ID
        """
        self.execute(f'UPDATE Classes set COLOR="{color}" where uid=={classId}')

    def listOfSlides(self):
        self.execute('SELECT uid,filename from Slides')
        return self.fetchall()

    def listOfSlidesWithExact(self) -> list:
        self.execute('SELECT uid,filename,exactImageID,directory from Slides')
        return self.fetchall()

    def getExactPerson(self) -> (int,str):
        ret = self.execute('SELECT uid,name from Persons where isExactUser=1 LIMIT 1').fetchone()
        if ret is None:
            return 0, ''
        else:
            return ret[0],ret[1]

    def setExactPerson(self, uid):
        self.execute(f'UPDATE Persons set isExactUser=0 where uid != {uid}')
        self.execute(f'UPDATE Persons set isExactUser=1 where uid == {uid}')
        self.commit()
        self.exactUser=uid

    def updateViewingProfile(self, vp:ViewingProfile):
        classes = self.getAllClasses()
        vp.COLORS_CLASSES = { 0: [0,0,0,0] }
        for name,id,color,clickable in classes:
            vp.COLORS_CLASSES[id] = [*hex_to_rgb(color),255]
        return vp

    def annotateImage(self, img: np.ndarray, leftUpper: list, rightLower:list, zoomLevel:float, vp : ViewingProfile, selectedAnnoID:int, annotationClasses):
        annos = self.getVisibleAnnotations(leftUpper, rightLower)
        self.VA = annos
        for idx,anno in annos.items():
            label = anno.agreedLabel()
            if (label in annotationClasses) and (annotationClasses[label]['Active']) and not anno.deleted:
                anno.draw(img, leftUpper, zoomLevel, thickness=2, vp=vp, selected=(selectedAnnoID==anno.uid))
    
    def findIntersectingAnnotation(self, anno:annotation, vp: ViewingProfile, database=None, annoType = None, annotationClasses = None):    
        if (database is None):
            database = self.VA            
        for idx,DBanno in database.items():
            if annotationClasses is not None and (DBanno.agreedLabel() in annotationClasses) and (annotationClasses[DBanno.agreedLabel()]['Active']) and not anno.deleted:
                if (DBanno.intersectingWithAnnotation(anno ))  and (DBanno.clickable):
                    if (annoType == DBanno.annotationType) or (annoType is None):
                        return DBanno
        return None

    def updateSlideFolder(self, slideUid:int, slidePath:str):
        directory = os.path.dirname(os.path.realpath(slidePath))
        self.execute(f'UPDATE Slides set directory="{directory}" where uid={slideUid}')

    def classPosition(self, classId):
        if (classId==0):
            return 0
        for pos, (name,uid,color) in enumerate(self._allclasses):
            if (uid==classId):
                return pos+1

    def findClickAnnotation(self, clickPosition, annotationClasses,  database=None, annoType = None, zoom:float=1.0):
        if (database is None):
            database = self.VA            
        for idx,anno in database.items():
            if (annotationClasses[0][anno.agreedLabel()]['Active']) and (annotationClasses[0][anno.agreedLabel()]['Clickable']):
                #if (vp.activeClasses[self.classPosition(anno.agreedLabel())]):
                if (anno.positionInAnnotation(clickPosition, zoom=zoom )) and (anno.clickable):
                    if (annoType == anno.annotationType) or (annoType is None):
                        return anno
        return None

    def getExactIDforSlide(self, slide):
        try:
            return self.execute(f'SELECT exactImageID from Slides where uid=={slide}').fetchone()[0]
        except:
            return ''



    def getAnnotationZLevels(self, slideId):
        """
            Gets all available zLevels (e.g., frame number) containing annotations 
        """
        if (slideId is None):
            return
        
        #self.dbcur.execute('SELECT uid, type,agreedClass,guid,lastModified,deleted,description FROM Annotations WHERE slide == %d'% slideId)
        self.dbcur.execute(f'SELECT coordinateZ FROM Annotations_coordinates where slide == {slideId} group by coordinateZ')
        return np.array(self.dbcur.fetchall()).reshape(-1).tolist()



    """

        Loads the current DB entry into memory. If the zLevel (for stack / time series images) is None, loads all annotations
    """

    def loadIntoMemory(self, slideId, transformer=None, zLevel=0):
        if not (self.isOpen()):
            return

        self.annotations = dict()
        self.annotationsSlide = slideId
        self.guids = dict()
        self.transformer = transformer
        self.zLevel = zLevel

        if (slideId is None):
            return

        self.dbcur.execute('SELECT uid, type,agreedClass,guid,lastModified,deleted,description,clickable FROM Annotations WHERE slide == %d'% slideId)
        allAnnos = self.dbcur.fetchall()


        self.dbcur.execute('SELECT coordinateX, coordinateY, coordinateZ, annoid FROM Annotations_coordinates where annoId IN (SELECT uid FROM Annotations WHERE slide == %d) ORDER BY orderIdx' % (slideId))
        allCoords = np.asarray(self.dbcur.fetchall())

        if (self.transformer is not None) and len(allCoords.shape)>1:
            allCoords = self.transformer(allCoords)

        for uid, annotype,agreedClass,guid,lastModified,deleted,description, clickable in allAnnos:
            coords = allCoords[allCoords[:,3]==uid,0:2]
            zCoord = allCoords[allCoords[:,3]==uid,2][0] if allCoords.shape[0]>0 else 0

            self.guids[guid] = uid
            if (zCoord != zLevel) and (zLevel is not None):
                continue
            if (annotype == AnnotationType.SPOT):
                self.annotations[uid] = spotAnnotation(uid, coords[0][0], coords[0][1], text=description)
            elif (annotype == AnnotationType.SPECIAL_SPOT):
                self.annotations[uid] = spotAnnotation(uid, coords[0][0], coords[0][1], True, text=description)
            elif (annotype == AnnotationType.POLYGON):
                self.annotations[uid] = polygonAnnotation(uid, coords, text=description)
            elif (annotype == AnnotationType.AREA):
                self.annotations[uid] = rectangularAnnotation(uid, coords[0][0], coords[0][1], coords[1][0], coords[1][1], text=description)
            elif (annotype == AnnotationType.CIRCLE):
                self.annotations[uid] = circleAnnotation(uid, coords[0][0], coords[0][1], coords[1][0], coords[1][1], text=description)
            elif (annotype == AnnotationType.IMAGEANNOTATION):
                self.annotations[uid] = imageAnnotation(uid, text=description)
            else:
                print('Unknown annotation type %d found :( ' % annotype)
            self.annotations[uid].agreedClass = agreedClass
            self.annotations[uid].guid = guid
            self.annotations[uid].zLevel = zCoord
            self.annotations[uid].lastModified = lastModified
            self.annotations[uid].deleted = deleted
            self.annotations[uid].clickable = clickable
            if (deleted):
                self.annotations[uid].clickable = False
        # Add all labels
        self.dbcur.execute('SELECT annoid, person, class,uid, exact_id FROM Annotations_label WHERE annoID in (SELECT uid FROM Annotations WHERE slide == %d)'% slideId)
        allLabels = self.dbcur.fetchall()

        for (annoId, person, classId,uid, exact_id) in allLabels:
            if (annoId) in self.annotations:
                self.annotations[annoId].addLabel(AnnotationLabel(person, classId, uid, exact_id), updateAgreed=False)


        self.generateMinMaxCoordsList()

            
    

    def checkTableStructure(self, tableName, action='check'):
        self.dbcur.execute('PRAGMA table_info(%s)' % tableName)
        ti = self.dbcur.fetchall()
        if (action=='check'):
            return self.databaseStructure[tableName].checkTableInfo(ti)
        elif (action=='ammend'):
            return self.databaseStructure[tableName].addMissingTableInfo(ti, tableName)

    

    def open(self, dbfilename):
        if os.path.isfile(dbfilename):
            self.db = sqlite3.connect(dbfilename)
            self.dbcur = self.db.cursor()
            self.db.create_function("generate_uuid",0, generate_uuid)
            self.db.create_function("pycurrent_time",0, time.time)
            self.db.create_function("randcolor",0, random_color)

            # Check structure of database and ammend if not proper
            if not self.checkTableStructure('Slides'):
                sql_statements = self.checkTableStructure('Slides','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Persons'):
                sql_statements = self.checkTableStructure('Persons','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Classes'):
                sql_statements = self.checkTableStructure('Classes','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Annotations'):
                sql_statements = self.checkTableStructure('Annotations','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Annotations_coordinates'):
                sql_statements = self.checkTableStructure('Annotations_coordinates','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Annotations_label'):
                sql_statements = self.checkTableStructure('Annotations_label','ammend')
                for sql in sql_statements:
                    self.dbcur.execute(sql)
                self.db.commit()

            if not self.checkTableStructure('Log'):
                # add new log, no problemo.
                self.dbcur.execute('DROP TABLE if exists `Log`')
                self.dbcur.execute(self.databaseStructure['Log'].getCreateStatement())
                self.db.commit()
            
            self.dbOpened = True
            # Migrating vom V0 to V1 needs proper filling of GUIDs
            DBversion = self.dbcur.execute('PRAGMA user_version').fetchone()
            if (DBversion[0]==0):
                self.dbcur.execute('UPDATE Annotations set guid=generate_uuid() where guid is NULL')

                self.addTriggers()

                # Add last polygon point (close the line)
                allpolys = self.dbcur.execute('SELECT uid from Annotations where type==3').fetchall()
                for [polyid,] in allpolys:
                    coords = self.dbcur.execute(f'SELECT coordinateX, coordinateY,slide FROM Annotations_coordinates where annoid=={polyid} and orderidx==1').fetchone()
                    maxidx = self.dbcur.execute(f'SELECT MAX(orderidx) FROM Annotations_coordinates where annoid=={polyid}').fetchone()[0]+1
                    self.dbcur.execute(f'INSERT INTO Annotations_coordinates (coordinateX, coordinateY, slide, annoId, orderIdx) VALUES ({coords[0]},{coords[1]},{coords[2]},{polyid},{maxidx} )')

                DBversion = self.dbcur.execute('PRAGMA user_version = 1')
                print('Successfully migrated DB to version 1')
                self.commit()
                DBversion = self.dbcur.execute('PRAGMA user_version').fetchone()

            if (DBversion[0]==1):
                self.dbcur.execute('UPDATE Classes set color=randcolor() where color is NULL')

                # changed Slides.directory to absolute name
                allslides = self.dbcur.execute('SELECT uid, filename, directory FROM Slides').fetchall()                
                for (uid, filename, pathname) in allslides:
                    pathname = '' if pathname is None else pathname # avoid None problems
                    filename = '' if filename is None else filename
                    dirname = '.' if os.path.dirname(dbfilename) is None else os.path.dirname(dbfilename)
                    candidates = [filename, 
                                  dirname+os.sep+filename,
                                  pathname+os.sep+filename,
                                  dbfilename+os.sep+pathname+os.sep+filename]
                    found=False
                    for cand in candidates:
                        if os.path.exists(cand):
                            fn=filename
                            found=True

                    if found:
                        self.updateSlideFolder(uid, fn)

                self.addTriggers()

                self.dbcur.execute('PRAGMA user_version = 2')
                print('Successfully migrated DB to version 2')
                self.commit()
                DBversion = self.dbcur.execute('PRAGMA user_version').fetchone()

            if (DBversion[0]==2):
                self.dbcur.execute('PRAGMA user_version = 3')
                print('Successfully migrated DB to version 3')
                self.commit()
                DBversion = self.dbcur.execute('PRAGMA user_version').fetchone()


            self.dbOpened = True
            self.dbfilename = dbfilename
            self.dbname = os.path.basename(dbfilename)
            self.getAllClasses()


            return self
        else:
            return False
    

    def deleteTriggers(self):
        for event in ['UPDATE','INSERT']:
            self.dbcur.execute(f'DROP TRIGGER IF EXISTS updateAnnotation_fromLabel{event}')
            self.dbcur.execute(f'DROP TRIGGER IF EXISTS updateAnnotation_fromCoords{event}')
            self.dbcur.execute(f'DROP TRIGGER IF EXISTS updateAnnotation_{event}')
            self.dbcur.execute(f'DROP TRIGGER IF EXISTS updateAnnotation_ins')

    def addTriggers(self):
        for event in ['UPDATE','INSERT']:
            self.dbcur.execute(f"""CREATE TRIGGER IF NOT EXISTS updateAnnotation_fromLabel{event}
                        AFTER {event} ON Annotations_label
                        BEGIN
                            UPDATE Annotations SET lastModified=pycurrent_time() where uid==new.annoId;
                        END
                        ;
                        """)

            self.dbcur.execute(f"""CREATE TRIGGER IF NOT EXISTS updateAnnotation_fromCoords{event}
                        AFTER {event} ON Annotations_coordinates
                        BEGIN
                            UPDATE Annotations SET lastModified=pycurrent_time() where uid==new.annoId;
                        END
                        ;
                        """)
            self.dbcur.execute(f"""CREATE TRIGGER IF NOT EXISTS updateAnnotation_{event}
                        AFTER {event} ON Annotations
                        BEGIN
                            UPDATE Annotations SET lastModified=pycurrent_time() where uid==new.uid;
                        END
                        ;
                        """)

        self.dbcur.execute(f"""CREATE TRIGGER IF NOT EXISTS updateAnnotation_ins
                    AFTER INSERT ON Annotations
                    BEGIN
                        UPDATE Annotations SET guid=generate_uuid() where uid==new.uid and guid is Null;
                    END
                    ;
                    """)

        self.dbcur.execute(f"""CREATE TRIGGER IF NOT EXISTS setRandomcolor
                    AFTER INSERT ON Classes
                    BEGIN
                        UPDATE Classes SET color=randcolor() where uid==new.uid and color is Null;
                    END
                    ;
                    """)
    
    # copy database to new file
    def saveTo(self, dbfilename):
        new_db = sqlite3.connect(dbfilename) # create a memory database

        query = "".join(line for line in self.db.iterdump())

        # Dump old database in the new one. 
        new_db.executescript(query)

        return True



    def getUnknownInCurrentScreen(self,leftUpper, rightLower, currentAnnotator) -> annotation:
        visAnnos = self.getVisibleAnnotations(leftUpper, rightLower)
        for anno in visAnnos.keys():
            if (visAnnos[anno].labelBy(currentAnnotator) == 0):
                return anno   
        return None

    def findAllAnnotationLabels(self, uid):
        q = 'SELECT person, class, uid FROM Annotations_label WHERE annoId== %d' % uid
        self.execute(q)
        return self.fetchall()

    def checkIfAnnotatorLabeled(self, uid, person):
        q = 'SELECT COUNT(*) FROM Annotations_label WHERE annoId== %d AND person == %d' % (uid, person)
        self.execute(q)
        return self.fetchone()[0]

    def findClassidOfClass(self, classname):
        q = 'SELECT uid FROM Classes WHERE name == "%s"' % classname
        self.execute(q)
        return self.fetchall()

    def findAllAnnotations(self, annoId, slideUID = None, zLevel=0):
        if (slideUID is None):
            q = 'SELECT coordinateX, coordinateY FROM Annotations_coordinates WHERE annoId==%d' % annoId
        else:
            q = 'SELECT coordinateX, coordinateY FROM Annotations_coordinates WHERE annoId==%d AND slide == %d and coordinateZ == %d' % (annoId, slideUID, zLevel)
        self.execute(q)
        return self.fetchall()    

    def findSlideForAnnotation(self, annoId):
        q = 'SELECT filename FROM Annotations_coordinates LEFT JOIN Slides on Slides.uid == Annotations_coordinates.slide WHERE annoId==%d' % annoId
        self.execute(q)
        return self.fetchall()    


    def pickRandomUnlabeled(self, byAnnotator=0) -> annotation:
        annoIds = list(self.annotations.keys())
        random.shuffle(annoIds)
        for annoId in annoIds:
            if (self.annotations[annoId].labelBy(byAnnotator) == 0):
                return self.annotations[annoId]
        return None



    def setSlideDimensions(self,slideuid,dimensions):
        if dimensions is None:
            return
        if (slideuid is None):
            return
        if (dimensions[0] is None):
            return
        if (dimensions[1] is None):
            return
        print('Setting dimensions of slide ',slideuid,'to',dimensions)
        self.execute('UPDATE Slides set width=%d, height=%d WHERE uid=%d' % (dimensions[0],dimensions[1],slideuid))
        self.db.commit()

    def slideFilenameForID(self, slideuid):
        try:
            return self.execute(f'SELECT filename FROM Slides where uid=={slideuid}').fetchone()[0]
        except:
            return ''

    def setPathForSlide(self, slideuid, slidepath):
        self.execute(f'UPDATE SLIDES set directory="{os.path.dirname(slidepath)}" where uid=={slideuid}')
        self.commit()
        return slideuid

    def findSlideWithFilename(self,slidename,slidepath, uuid:str=None):
        if (slidepath=='') and len(os.path.split(slidename)[0])>0:
            slidepath,slidename = os.path.split(slidename)

        if (len(os.path.split(slidepath)[0])>1):
            directory,fname = os.path.split(slidepath)
        else:
            directory = slidepath

        ret = self.execute('SELECT uid,directory,uuid,filename from Slides ').fetchall()
        #print('All slides:',ret)
        #print('Looking up',slidename,'in',directory)
        secondBest=None
        for (uid,slidedir,suuid,fname) in ret:
            if (uuid is not None) and (suuid==uuid):
                return uid
            elif (fname==slidename) or (os.path.basename(fname)==slidename):
                if slidedir is None:
                    secondBest=uid
                elif (slidedir.upper() == directory.upper()):
                    return uid 
                elif (os.path.realpath(slidedir).upper() == directory.upper()):
                    return uid
                else:
                    secondBest=uid
        return secondBest
    
    def insertAnnotator(self, name):
        self.execute('INSERT INTO Persons (name) VALUES ("%s")' % (name))
        self.commit()
        query = 'SELECT last_insert_rowid()'
        return self.execute(query).fetchone()[0]

    def changeAllAnnotationLabelsOfType(self, classId:int, annotationType:int, newClassId:int):
        self.execute(f'UPDATE Annotations_label set class={newClassId} where annoid in (SELECT uid from Annotations where type=={annotationType}) and class=={classId}')
        self.execute(f'UPDATE Annotations set agreedClass={newClassId} where uid in (SELECT uid from Annotations where type=={annotationType}) and agreedClass=={classId}')

    def insertClass(self, name):
        self.execute('INSERT INTO Classes (name) VALUES ("%s")' % (name))
        self.commit()
        query = 'SELECT last_insert_rowid()'
        return self.execute(query).fetchone()[0]
    
    def setAgreedClass(self, classId, annoIdx):
        self.annotations[annoIdx].agreedClass = classId
        q = 'UPDATE Annotations SET agreedClass==%d WHERE uid== %d' % (classId,annoIdx)
        self.execute(q)
        self.commit()
         

    def setAnnotationLabel(self,classId,  person, entryId, annoIdx, **kwargs):
        if ('exact_id' in kwargs): # try to keep exact_id, if not explicitly overwritten
            q = f'UPDATE Annotations_label SET person=={person}, class={classId}, exact_id={kwargs["exact_id"]} WHERE uid== {entryId}'
        else:
            q = f'UPDATE Annotations_label SET person=={person}, class={classId} WHERE uid== {entryId}'
            
        self.execute(q)
        self.commit()
        self.annotations[annoIdx].changeLabel(entryId, person, classId)
        
        # check if all labels belong to one class now
        if np.all(np.array([lab.classId for lab in self.annotations[annoIdx].labels])==classId):
            self.setAgreedClass(classId, annoIdx)

        self.checkCommonAnnotation(annoIdx)

    def checkCommonAnnotation(self, annoIdx ):
        allAnnos = self.findAllAnnotationLabels(annoIdx)
        if (len(allAnnos)>0):
            matching = allAnnos[0][1]
            for anno in allAnnos:
                if (anno[1] != matching):
                    matching=0
        else:
            matching=0
        
        q = 'UPDATE Annotations set agreedClass=%d WHERE UID=%d' % (matching, annoIdx)
        self.execute(q)
        self.commit()

    def logLabel(self, labelId):
        query = 'INSERT INTO Log (dateTime, labelId) VALUES (%d, %d)' % (time.time(), labelId)
        self.execute(query)


    def addAnnotationLabel(self,classId,  person, annoId, exact_id:int=None):
        if (exact_id is None):
            exact_id = 'Null'
        query = ('INSERT INTO Annotations_label (person, class, annoId, exact_id) '
                  f'VALUES ({person},{classId},{annoId},{exact_id})')
        self.execute(query)
        query = 'SELECT last_insert_rowid()'
        self.execute(query)
        newid = self.fetchone()[0]
        self.logLabel(newid)

        if annoId in self.annotations: # slide needs to be loaded in order to work
            self.annotations[annoId].addLabel(AnnotationLabel(person, classId, newid, exact_id))

        self.checkCommonAnnotation( annoId)
        self.commit()

    def exchangePolygonCoordinates(self, annoId, slideUID, annoList, zLevel):
        self.annotations[annoId].annotationType = AnnotationType.POLYGON
        self.annotations[annoId].coordinates = np.asarray(annoList)
        self.generateMinMaxCoordsList()

        query = 'DELETE FROM Annotations_coordinates where annoId == %d' % annoId
        self.execute(query)
        self.commit()

        self.insertCoordinates(np.array(annoList), slideUID, annoId, zLevel)
        

    def insertCoordinates(self, annoList:np.ndarray, slideUID, annoId, zLevel):
        """
                Insert an annotation into the database.
                annoList must be a numpy array, but can be either 2 columns or 3 columns (3rd is order)
        """

        if self.transformer is not None:
            annoList = self.transformer(annoList, inverse=True)

        for num, annotation in enumerate(annoList.tolist()):
            query = (f'INSERT INTO Annotations_coordinates (coordinateX, coordinateY, coordinateZ, slide, annoId, orderIdx) VALUES ({annotation[0]},{annotation[1]},{zLevel}, {slideUID},{annoId},{annotation[2] if len(annotation)>2 else num+1})')
            self.execute(query)
        self.commit()

    def setPolygonCoordinates(self, annoId, coords, slideUID, zLevel):
        self.execute(f'DELETE FROM Annotations_coordinates where annoId={annoId}')

        self.insertCoordinates(np.array(coords), slideUID, annoId, zLevel)
        self.commit()

    def insertNewPolygonAnnotation(self, annoList, slideUID, classID, annotator, closed:bool=True, exact_id="Null", description:str='', zLevel:int=0):
        query = 'INSERT INTO Annotations (slide, agreedClass, type, description) VALUES (%d,%d,3,"%s")' % (slideUID,classID, description)
#        query = 'INSERT INTO Annotations (coordinateX1, coordinateY1, coordinateX2, coordinateY2, slide, class1, person1) VALUES (%d,%d,%d,%d,%d,%d, %d)' % (x1,y1,x2,y2,slideUID,classID,annotator)
        if (isinstance(annoList, np.ndarray)):
            annoList=annoList.tolist()
        if (closed):
            annoList.append(annoList[0])
        self.execute(query)
        query = 'SELECT last_insert_rowid()'
        self.execute(query)
        annoId = self.fetchone()[0]
        assert(len(annoList)>0)
        self.annotations[annoId] = polygonAnnotation(annoId, np.asarray(annoList))
        self.appendToMinMaxCoordsList(self.annotations[annoId])
        self.insertCoordinates(np.array(annoList), slideUID, annoId, zLevel)

        self.addAnnotationLabel(classId=classID, person=annotator, annoId=annoId, exact_id=exact_id)

        self.commit()
        return annoId

    def setAnnotationClickable(self, annoId:int, clickable:int):

        self.execute(f'UPDATE ANNOTATIONS SET clickable={clickable} where uid=={annoId}')

    def addAnnotationToDatabase(self, anno:annotation, slideUID:int, classID:int, annotatorID:int, zLevel:int, description:str='') -> int:
        if (anno.annotationType == AnnotationType.AREA):
            return self.insertNewAreaAnnotation(anno.x1,anno.y1,anno.x2,anno.y2,slideUID,classID, annotatorID,description=description, zLevel=zLevel)
        elif (anno.annotationType == AnnotationType.POLYGON):
            return self.insertNewPolygonAnnotation(anno.coordinates, slideUID, classID, annotatorID, description=description, zLevel=zLevel)
        elif (anno.annotationType == AnnotationType.CIRCLE):
            coords = anno.coordinates
            return self.insertNewAreaAnnotation(coords[0],coords[1],coords[2],coords[3],slideUID,classID, annotatorID, typeId=5,description=description, zLevel=zLevel)
        elif (anno.annotationType == AnnotationType.SPOT):
            return self.insertNewSpotAnnotation(anno.x1, anno.y1, slideUID, classID, annotatorID, description=description, zLevel=zLevel)
        elif (anno.annotationType == AnnotationType.SPECIAL_SPOT):
            return self.insertNewSpotAnnotation(anno.x1, anno.y1, slideUID, classID, annotatorID, type=4, description=description, zLevel=zLevel)
        

    def getGUID(self, annoid) -> str:
        try:
            return self.execute(f'SELECT guid from Annotations where uid=={annoid}').fetchone()[0]
        except:
            return None

    def insertNewAreaAnnotation(self, x1,y1,x2,y2, slideUID, classID, annotator, typeId=2, uuid=None, exact_id="Null", description="", zLevel=0):
        if (uuid is None):
            query = 'INSERT INTO Annotations (slide, agreedClass, type, description) VALUES (%d,%d,%d,"%s")' % (slideUID,classID, typeId, description)
        else:
            query = 'INSERT INTO Annotations (slide, agreedClass, type, guid, description) VALUES (%d,%d,%d,"%s","%s")' % (slideUID,classID, typeId, uuid,description)
#        query = 'INSERT INTO Annotations (coordinateX1, coordinateY1, coordinateX2, coordinateY2, slide, class1, person1) VALUES (%d,%d,%d,%d,%d,%d, %d)' % (x1,y1,x2,y2,slideUID,classID,annotator)
        self.execute(query)
        query = 'SELECT last_insert_rowid()'
        self.execute(query)
        annoId = self.fetchone()[0]
        if (typeId==2):
            self.annotations[annoId] = rectangularAnnotation(annoId, x1,y1,x2,y2)
        else:
            self.annotations[annoId] = circleAnnotation(annoId, x1,y1,x2,y2)
        
        self.appendToMinMaxCoordsList(self.annotations[annoId])

        annoList = np.array([[x1,y1,1],[x2,y2,2]])
        self.insertCoordinates(np.array(annoList), slideUID, annoId, zLevel)
        
        self.addAnnotationLabel(classId=classID, person=annotator, annoId=annoId, exact_id=exact_id)

        self.commit()
        return annoId

    def setLastModified(self, annoid:int, lastModified:float):
        self.execute(f'UPDATE Annotations SET lastModified="{lastModified}" where uid={annoid}')
        self.annotations[annoid].lastModified = lastModified

    def setGUID(self, annoid:int, guid:str):
        self.execute(f'UPDATE Annotations SET guid="{guid}" where uid={annoid}')
        self.guids[guid]=annoid
    
    def removeImageAnnotation(self, slideUID:int, zLevel:int, annotator:int, exact_id="Null"):
        query = f'SELECT annoId FROM Annotations_coordinates where coordinateX IS NULL and coordinateY IS NULL and coordinateZ=={zLevel} and slide=={slideUID} and annoID in (SELECT annoID FROM Annotations_label where person=={annotator})'
        # delete all previous labels by current annotator
        self.execute(query)
        for [aid,] in self.fetchall():
            self.removeAnnotation(aid)


    def insertNewImageAnnotation(self, slideUID:int, zLevel:int, classID:int, annotator:int, exact_id="Null", description:str=''):

        self.removeImageAnnotation(slideUID=slideUID, zLevel=zLevel, annotator=annotator)

        query = 'INSERT INTO Annotations (slide,agreedClass,type,description) VALUES (%d,0,%d, "%s")' % (slideUID, AnnotationType.IMAGEANNOTATION,description)
        self.execute(query)
        query = 'SELECT last_insert_rowid()'
        self.execute(query)
        annoId = self.fetchone()[0]
        self.annotations[annoId] = imageAnnotation(annoId)
        self.insertCoordinates(np.array([['Null', 'Null',1]]), slideUID, annoId, zLevel)
        self.addAnnotationLabel(classId=classID, person=annotator, annoId=annoId, exact_id=exact_id)
        self.appendToMinMaxCoordsList(self.annotations[annoId])



    def insertNewSpotAnnotation(self,xpos_orig,ypos_orig, slideUID, classID, annotator, type = 1, exact_id="Null", description:str='', zLevel=0):

        if (type == 4):
            query = 'INSERT INTO Annotations (slide, agreedClass, type,description) VALUES (%d,0,%d, "%s")' % (slideUID, type,description)
            self.execute(query)
            query = 'SELECT last_insert_rowid()'
            self.execute(query)
            annoId = self.fetchone()[0]

            self.insertCoordinates(np.array([[xpos_orig, ypos_orig,1]]), slideUID, annoId, zLevel)
            self.annotations[annoId] = spotAnnotation(annoId, xpos_orig,ypos_orig, (type==4))

        else:
            query = 'INSERT INTO Annotations (slide, agreedClass, type, description) VALUES (%d,%d,%d,"%s")' % (slideUID,classID, type,description)
            self.execute(query)
            query = 'SELECT last_insert_rowid()'
            self.execute(query)
            annoId = self.fetchone()[0]

            self.insertCoordinates(np.array([[xpos_orig, ypos_orig,1]]), slideUID, annoId, zLevel)
            self.execute(query)

            self.annotations[annoId] = spotAnnotation(annoId, xpos_orig,ypos_orig, (type==4))
            self.addAnnotationLabel(classId=classID, person=annotator, annoId=annoId, exact_id=exact_id)

        self.appendToMinMaxCoordsList(self.annotations[annoId])
        self.commit()
        return annoId

    def removeFileFromDatabase(self, fileUID:str):
        self.execute('DELETE FROM Annotations_label where annoID IN (SELECT uid FROM Annotations where slide == %d)' % fileUID)
        self.execute('DELETE FROM Annotations_coordinates where annoID IN (SELECT uid FROM Annotations where slide == %d)' % fileUID)
        self.execute('DELETE FROM Annotations where slide == %d' % fileUID)
        self.execute('DELETE FROM Slides where uid == %d' % fileUID)
        self.commit()

    def removeAnnotationLabel(self, labelIdx, annoIdx):
            q = 'DELETE FROM Annotations_label WHERE uid == %d' % labelIdx
            self.execute(q)
            self.annotations[annoIdx].removeLabel(labelIdx)
            self.commit()
            self.checkCommonAnnotation(annoIdx)

    def updatePolygonPoint(self, annoId, orderIdx, coords):
        self.execute(f'UPDATE Annotations_coordinates SET coordinateX={coords[0]}, coordinateY={coords[1]} where annoId=={annoId} and orderIdx=={orderIdx}+1')
        self.commit()

    def removePolygonPoint(self, annoId:int, coord_idx:int):
        self.execute(f'DELETE FROM Annotations_coordinates where annoId=={annoId} and orderIdx=={coord_idx}+1')
        self.execute(f'UPDATE Annotations_coordinates SET orderidx=orderidx-1 where annoId=={annoId} and orderIdx>{coord_idx}')
        retain = [np.arange(0,coord_idx).tolist()+np.arange(coord_idx+1,self.annotations[annoId].coordinates.shape[0]).tolist()]
        self.annotations[annoId].coordinates = self.annotations[annoId].coordinates[retain,:][0,:,:]
        self.commit()

    def changeAnnotationID(self, annoId:int, newAnnoID:int):
            self.execute('SELECT COUNT(*) FROM Annotations where uid == (%d)' % newAnnoID)
            if (self.fetchone()[0] > 0):
                return False

            self.execute('UPDATE Annotations_label SET annoId = %d WHERE annoId == %d' % (newAnnoID,annoId))
            self.execute('UPDATE Annotations_coordinates SET annoId = %d WHERE annoId == %d' % (newAnnoID,annoId))
            self.execute('UPDATE Annotations SET uid= %d WHERE uid == %d ' % (newAnnoID,annoId))
            self.annotations[annoId].uid = newAnnoID
            self.annotations[newAnnoID] = self.annotations[annoId]
            self.annotations.pop(annoId)
            self.appendToMinMaxCoordsList(self.annotations[newAnnoID])
            self.commit()
            return True


    def removeAnnotation(self, annoId, onlyMarkDeleted:bool=True):
            if (onlyMarkDeleted):
                self.execute('UPDATE Annotations SET deleted=1 WHERE uid == '+str(annoId))
                self.annotations[annoId].deleted = 1
                self.annotations[annoId].clickable = 0
            else:
                self.execute('DELETE FROM Annotations_label WHERE annoId == %d' % annoId)
                self.execute('DELETE FROM Annotations_coordinates WHERE annoId == %d' % annoId)
                self.execute('DELETE FROM Annotations WHERE uid == '+str(annoId))
                self.annotations.pop(annoId)
            self.commit()
            

    def insertNewSlide(self,slidename:str,slidepath:str,uuid:str=""):
            
            # Should some put a path into the slide name, use that as a path
            if (len(os.path.split(slidename)[0])>0):
                if (slidepath==''):
                    slidepath = os.path.dirname(slidename)
                slidename = os.path.basename(slidename) 
 
            # only use the last part of the path
            if (len(os.path.split(slidepath)[0])>0):
                directory = os.path.dirname(os.path.realpath(slidepath))
            else:
                directory = ''

            self.execute('INSERT INTO Slides (filename,directory,uuid) VALUES ("%s","%s", "%s")' % (slidename,directory,uuid))
            self.commit()
            query = 'SELECT last_insert_rowid()'
            return self.execute(query).fetchone()[0]


    def fetchall(self):
        if (self.isOpen()):
            return self.dbcur.fetchall()
        else:
            print('Warning: DB not opened for fetch.')

    def fetchone(self):
        if (self.isOpen()):
            return self.dbcur.fetchone()
        else:
            print('Warning: DB not opened for fetch.')

    def execute(self, query):
        if (self.isOpen()):
            try:
                return self.dbcur.execute(query)
            except Exception as e:
                raise type(e)(str(e) +
                      ' Failed query was: %s' % query).with_traceback(sys.exc_info()[2])
        else:
            print('Warning: DB not opened.')

    def getDBname(self):
        if (self.dbOpened):
            return self.dbname
        else:
            return ''
    
    def getAnnotatorByID(self, id):
        if (id is None):
            return ''
        self.execute('SELECT name FROM Persons WHERE uid == %d' % id)
        fo = self.fetchone()
        if (fo is not None):
            return fo[0]
        else:
            return ''

    def getClassByID(self, id):
        try:
            self.execute('SELECT name FROM Classes WHERE uid == %d' % id)
            return self.fetchone()[0]
        except:
            return '-unknown-'+str(id)+'-'


    def getAllPersons(self):
        self.execute('SELECT name, uid FROM Persons ORDER BY uid')
        return self.fetchall()

    def getAllClasses(self):
        self.execute('SELECT name,uid,color,clickable FROM Classes ORDER BY uid')
        self._allclasses = self.fetchall() # save for later
        return self._allclasses
    
    def renameClass(self, classID, name):
        self.execute('UPDATE Classes set name="%s" WHERE uid ==  %d' % (name, classID))
        self.commit()

    def setClassClickable(self, classID, clickable):
        self.execute('UPDATE Classes set clickable=%d where uid == %d' % (clickable, classID))
        self.commit()

    def deleteClass(self, classID):
        self.execute('DELETE FROM Classes WHERE uid ==  %d' % ( classID))
        self.execute('UPDATE Annotations set agreedClass=0 WHERE agreedClass ==  %d' % ( classID))
        self.execute('DELETE FROM Annotations_label WHERE class ==  %d' % ( classID))
        self.commit()

    def commit(self):
        return self.db.commit()

    def countEntryPerClass(self, slideID = 0):
        retval = {'unknown' :  {'uid': 0, 'count_total':0, 'count_slide':0}}
        self.dbcur.execute('SELECT Classes.uid, COUNT(*), name FROM Annotations LEFT JOIN Classes on Classes.uid == Annotations.agreedClass where Annotations.deleted == 0 GROUP BY Classes.uid')
        allClasses = self.dbcur.fetchall()
    
        classids = np.zeros(len(allClasses))        
        for idx,element in enumerate(allClasses):
                name = element[2] if element[2] is not None else 'unknown'
                
                retval[name] = {'uid': element[0], 'count_total':element[1], 'count_slide':0}
        uidToName = {uid:name for uid,_,name in allClasses}


        if (slideID is not None):

            self.dbcur.execute('SELECT Classes.uid, COUNT(*), name FROM Annotations LEFT JOIN Classes on Classes.uid == Annotations.agreedClass where slide==%d and deleted==0 GROUP BY Classes.uid' % slideID)
            allClasses = self.dbcur.fetchall()
            
            for uid,cnt,name in allClasses:
                name = 'unknown' if name is None else name
                if name not in retval:
                    retval[name] = {'uid': 0, 'count_total':0, 'count_slide':0}
                retval[name]['count_slide'] = cnt

        return retval


    def countEntries(self):
        self.dbcur.execute('SELECT COUNT(*) FROM Annotations where deleted==0')
        num1 = self.dbcur.fetchone()

        return num1[0]

    def fetchSpotAnnotation(self,entryId=0):
        self.dbcur.execute('SELECT coordinateX, coordinateY FROM Annotations_coordinates WHERE annoId ==  '+str(entryId))
        coords = self.dbcur.fetchone()
        
        class1,class2,person1,person2=None,None,None,None
        self.dbcur.execute('SELECT Classes.name FROM Annotations_label LEFT JOIN Classes on Annotations_label.class == Classes.uid WHERE Annotations_label.annoId ==  '+str(entryId))
        classes = self.dbcur.fetchall()

        self.dbcur.execute('SELECT Persons.name FROM Annotations_label LEFT JOIN Persons on Annotations_label.person == Persons.uid WHERE Annotations_label.annoId ==  '+str(entryId))
        persons = self.dbcur.fetchall()

        return coords, classes, persons

    def fetchAreaAnnotation(self,entryId=0):
        self.dbcur.execute('SELECT coordinateX, coordinateY FROM Annotations_coordinates WHERE annoId ==  '+str(entryId)+' ORDER BY Annotations_coordinates.orderIdx')
        coords1 = self.dbcur.fetchone()
        coords2 = self.dbcur.fetchone()
        
        class1,class2,person1,person2=None,None,None,None
        self.dbcur.execute('SELECT Classes.name FROM Annotations_label LEFT JOIN Classes on Annotations_label.class == Classes.uid WHERE Annotations_label.annoId ==  '+str(entryId))
        classes = self.dbcur.fetchall()

        self.dbcur.execute('SELECT Persons.name FROM Annotations_label LEFT JOIN Persons on Annotations_label.person == Persons.uid WHERE Annotations_label.annoId ==  '+str(entryId))
        persons = self.dbcur.fetchall()

        return coords1, coords2, classes, persons

    

    def create(self,dbfilename) -> bool:
 
        if (os.path.isfile(dbfilename)):
             # ok, remove old file
            os.remove(dbfilename)

        try:
            tempdb = sqlite3.connect(dbfilename)
        except sqlite3.OperationalError:
            return False

        tempcur = tempdb.cursor()
        tempdb.create_function("randcolor",0, random_color)

        tempcur.execute('CREATE TABLE `Annotations_label` ('
            '	`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '	`person`	INTEGER,'
            '	`class`	INTEGER,'
            '   `exact_id` INTEGER,'
            '	`annoId`	INTEGER'
            ');')
        
        tempcur.execute('CREATE TABLE `Annotations_coordinates` ('
            '`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '`coordinateX`	INTEGER,'
            '`coordinateY`	INTEGER,'
            '`coordinateZ`	INTEGER DEFAULT 0,'
            '`slide`	INTEGER,'
            '`annoId`	INTEGER,'
            '`orderIdx`	INTEGER'
            ');')

        tempcur.execute('CREATE TABLE `Annotations` ('
            '`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '`slide`	INTEGER,'
           	'`guid`	TEXT,'
            f'`lastModified`	REAL DEFAULT {time.time()},'
         	'`deleted`	INTEGER DEFAULT 0,'
            '`type`	INTEGER,'
            '`agreedClass`	INTEGER,'
            '`description` TEXT,'
            '`clickable` INTEGER DEFAULT 1'
            ');')

        tempcur.execute('CREATE TABLE `Classes` ('
            '`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '`name`	TEXT,'
            '`color` TEXT'
            ');')
        
        tempcur.execute('CREATE TABLE `Persons` ('
            '`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '`name`	TEXT,'
            '`isExactUser` INTEGER DEFAULT 0'
            ');')

        tempcur.execute('CREATE TABLE `Slides` ('
            '`uid`	INTEGER PRIMARY KEY AUTOINCREMENT UNIQUE,'
            '`filename`	TEXT,'
            '`width`	INTEGER,'
            '`height`	INTEGER,'
            '`directory` TEXT,'
            '`uuid` TEXT,'
            '`exactImageID` TEXT'
            ');')

        tempdb.commit()
        self.db = tempdb
        self.dbcur = self.db.cursor()
        self.dbcur.execute('PRAGMA user_version = 1')
        self.db.create_function("generate_uuid",0, generate_uuid)
        self.db.create_function("pycurrent_time",0, time.time)
        self.dbfilename = dbfilename
        self.dbcur.execute(self.databaseStructure['Log'].getCreateStatement())
        self.annotations = dict()
        self.dbname = os.path.basename(dbfilename)
        self.dbOpened=True
        self.generateMinMaxCoordsList()
        self.addTriggers()

        return self



"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Support for CellVizio MKT files. 


"""

import numpy as np
from PIL import Image
import io
import os
import hashlib
import struct
from os import stat


class fileinfo:
    offset = 16 # fixed: 16 byte header
    gapBetweenImages = 32 # fixed: 32 byte gap between images
    size = 0 # size of file
    width = 0 # width of images
    height = 0 # height of images
    nImages = 0 # number of images
    circMask = 0; # circular mask (round shape)

class circularMask:
    mask = 0
    def __init__(self,w,h,r):
        self.x,self.y = np.ogrid[-int(h/2):np.ceil(h/2), -int(w/2):np.ceil(w/2)]
        self.mask = self.x*self.x + self.y*self.y <= int(r/2)*int(r/2)


class ReadableCellVizioMKTDataset():

    fi = fileinfo()

    def __init__(self, filename):

       self.fileName = filename
       self.isOpenSlide = False
       self.fileHandle = open(filename, 'rb')
       self.fileHandle.seek(5) # we find the FPS at position 05
       fFPSByte = self.fileHandle.read(4)
       self.fps = struct.unpack('>f', fFPSByte)[0]


       self.fileHandle.seek(10) # we find the image size at position 10
       fSizeByte = self.fileHandle.read(4)
       self.fi.size = int.from_bytes(fSizeByte, byteorder='big', signed=True)
       self.fi.nImages=1000

       # self.fi.width = 576
       # if ((self.fi.size/(2*self.fi.width))%2!=0):
       #      self.fi.width=512
       #      self.fi.height=int(self.fi.size/(2*self.fi.width))
       # else:
       #      self.fi.height=int(self.fi.size/(2*self.fi.width))
       self.fi.width = 576
       if ((self.fi.size / (2 * self.fi.width)) % 2 != 0):
           self.fi.width = 512
       self.fi.height = int(self.fi.size / (2 * self.fi.width))

       if abs(self.fi.width - self.fi.height) > 100:
           self.fi.width = 448
           self.fi.height = int(self.fi.size / (2 * self.fi.width))

       self.filestats = stat(self.fileName)
       self.fi.nImages = int((self.filestats.st_size-self.fi.offset) / (self.fi.size+self.fi.gapBetweenImages))

       self.fileHandle.seek(0) # we find the image size at position 10
       hashablecontent = self.fileHandle.read(1024)
       self.hash_sha = hashlib.sha256(hashablecontent).hexdigest()

       self.numberOfFrames = self.fi.nImages

       self.geometry_imsize = [self.fi.height, self.fi.width]
       self.imsize = [self.fi.height, self.fi.width]
       self.geometry_tilesize = [(self.fi.height, self.fi.width)]
       self.geometry_rows = [1]
       self.geometry_columns = [1]
       self.levels = [1]
       self.channels = 1
       self.mpp_x = 1e-6
       self.mpp_y  = 1e-6

       self.circMask = circularMask(self.fi.width,self.fi.height, self.fi.width-2).mask

    

    @property
    def seriesInstanceUID(self) -> str:
        return self.hash_sha

    @property
    def level_downsamples(self):
        return [1]

    @property 
    def level_dimensions(self):
        return [self.geometry_imsize]

    def readImage(self, position=0):

       self.fileHandle.seek(self.fi.offset + self.fi.size*position + self.fi.gapBetweenImages*position)

       image = np.fromfile(self.fileHandle, dtype=np.int16, count=int(self.fi.size/2))

       image=np.reshape(image, newshape=(self.fi.height, self.fi.width))

       image = np.ma.masked_array(image, 1-self.circMask) # apply mask to image
       return image

    def scaleImageUINT8(self, image, mask = None):
       # read image and scale to uint8 [0;255] format

       if (mask is None):
           mask = self.circMask

       maskedImage = image[mask]

       cmin,cmax = np.percentile(maskedImage,0.5), np.percentile(maskedImage,99.5)
       if (cmax>5000):
           cmax=5000
       dyn=cmax-cmin

       # compress
       compr=255/dyn
       image = image-cmin
       image = image*compr

       # limit to 0
       image = np.clip(np.round(image),0,255)
       image=np.uint8(image)

       return image
    def readImageUINT8(self, position=0):
       # read image and scale to uint8 [0;255] format
       image=self.readImage(position)

       image = self.scaleImageUINT8(image)
       return image

    def read_region(self, location: tuple, level:int, size:tuple, zLevel:int):
        img = np.zeros((size[1],size[0],4), np.uint8)
        img[:,:,3]=255
        offset=[0,0]
        if (location[1]<0):
            offset[0] = -location[1]
            location = (location[0],0)
        if (location[0]<0):
            offset[1] = -location[0]
            location = (0,location[1])

        pixel_array = self.readImageUINT8(position=zLevel)
        imgcut = pixel_array[location[1]:location[1]+size[1]-offset[0],location[0]:location[0]+size[0]-offset[1]]
        imgcut = np.uint8(np.clip(np.float32(imgcut),0,255))

        for k in range(3):
            img[offset[0]:imgcut.shape[0]+offset[0],offset[1]:offset[1]+imgcut.shape[1],k] = imgcut
        return Image.fromarray(img)


    @property
    def dimensions(self):
        return self.level_dimensions[0]

    def get_best_level_for_downsample(self,downsample):
        return np.argmin(np.abs(np.asarray(self.level_downsamples)-downsample))

    @property
    def level_count(self):
        return len(self.levels)

    def imagePos_to_id(self, imagePos:tuple, level:int):
        id_x, id_y = imagePos
        if (id_y>=self.geometry_rows[level]):
            id_x=self.geometry_columns[level] # out of range

        if (id_x>=self.geometry_columns[level]):
            id_y=self.geometry_rows[level] # out of range
        return (id_x+(id_y*self.geometry_columns[level]))
    

    def get_id(self, pixelX:int, pixelY:int, level:int) -> (int, int, int):

        id_x = round(-0.5+(pixelX/self.geometry_tilesize[level][1]))
        id_y = round(-0.5+(pixelY/self.geometry_tilesize[level][0]))
        
        return (id_x,id_y), pixelX-(id_x*self.geometry_tilesize[level][0]), pixelY-(id_y*self.geometry_tilesize[level][1]),

    



"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Support for DICOM WSI files. 

   To create a DICOM WSI file out of an openslide-supported WSI file, use wsi2dcm, available from:

   https://github.com/GoogleCloudPlatform/wsi-to-dicom-converter


"""

import numpy as np
try:
    import pydicom
    from pydicom.encaps import decode_data_sequence
    dic_fileformats = ['*.dcm']
except:
    dic_fileformats = []

from PIL import Image
import io
import os

class sequencedTiles(object):
  #
  def __init__(self, dsstore):
      self._dsstore = dsstore
      self._dsequence = dict()

  def __getitem__(self, key):
    if (key) in self._dsequence:
        return self._dsequence[key]
    elif key not in self._dsstore:
        raise ValueError('Entry does not exist')
    else:
        self._dsequence[key] = decode_data_sequence(self._dsstore[key].PixelData)
        return self._dsequence[key]
  #
  def __setitem__(self, key, value):
    raise NotImplementedError()

class ReadableDicomDataset():
    def __init__(self, filename):

        self._ds = pydicom.dcmread(filename)
        self._sequenceInstanceUID = self._ds.SeriesInstanceUID
        self.isOpenSlide=False
        
        self._path, _ = os.path.split(filename)
        if (self._path==""):
            self._path='.'+os.sep

        self._dsstore = dict()

        # look for all other levels
        for dcmfile in os.listdir(self._path):
            if os.path.splitext(dcmfile)[-1].upper()=='.DCM':
                try:
                    tmpdcm = pydicom.dcmread(self._path+os.sep+dcmfile)
                    if (tmpdcm.InstanceNumber) is None:
                        self._dsstore[0] = tmpdcm      
                    elif (tmpdcm.SeriesInstanceUID==self._sequenceInstanceUID):
                        self._dsstore[tmpdcm.InstanceNumber-1] = tmpdcm            
                except Exception as e:
                    print(f'Warning: Unable to open {dcmfile}:',e)
        
        # some DICOM files have no instance 0 (as DICOM WSIs)
        if 0 not in self._dsstore:
            self._dsstore = {0: list(self._dsstore.values())[0]}

        # check if all instances have same resolution --> this is a stack
        dims_instances = [[self._dsstore[x].Columns,self._dsstore[x].Rows ] for x in self._dsstore]
        stack = all([dims_instances[x][0]==dims_instances[0][0] for x in range(len(dims_instances))]) and \
                all([dims_instances[x][1]==dims_instances[0][1] for x in range(len(dims_instances))]) and len(dims_instances)>1    
        self.multiInstanceStack = stack

        self._dsequence = sequencedTiles(self._dsstore)
        
        try:
            self.levels = sorted(list(self._dsstore.keys()))
            self.geometry_imsize = [(self._dsstore[k][0x48,0x6].value,self._dsstore[k][0x48,0x7].value) for k in self.levels]
            self.geometry_tilesize = [(self._dsstore[k].Columns, self._dsstore[k].Rows) for k in self.levels]
            self.geometry_columns = [round(0.5+(self.geometry_imsize[k][0]/self.geometry_tilesize[k][0])) for k in self.levels]
            self.geometry_rows = [round(0.5 + (self.geometry_imsize[k][1] / self.geometry_tilesize[k][1] )) for k in self.levels]
            self.channels = self._ds[0x0028, 0x0002].value
            self.mpp_x = float(self._dsstore[0].SharedFunctionalGroupsSequence[0][0x028,0x9110][0][0x028,0x030][0])*1000
            self.mpp_y = float(self._dsstore[0].SharedFunctionalGroupsSequence[0][0x028,0x9110][0][0x028,0x030][1])*1000
            self.numberOfFrames = 1
            self.loadIntoMemory=False
        except Exception as e:
            # Fall back to simple reader
            self.levels = sorted(list(self._dsstore.keys())) if not stack else [0]
            self.loadIntoMemory=True
#            self.read_region = self.read_region_simple
            self.extremes = np.min(self._dsstore[0].pixel_array), np.max(self._dsstore[0].pixel_array)
            self.geometry_imsize = [self._dsstore[0].Columns, self._dsstore[0].Rows]
            self.geometry_tilesize = [(self._dsstore[0].Columns, self._dsstore[0].Rows)]
            self.geometry_rows = [1]
            self.geometry_columns = [1]
            self.channels = 1
            self.mpp_x = 1e-6
            self.mpp_y  = 1e-6

            if not stack:
                self.numberOfFrames = self._dsstore[0].NumberOfFrames if ('NumberOfFrames' in self._dsstore[0]) else 1
            else:
                self.numberOfFrames = len(dims_instances)
            self.fps = 1000/self._dsstore[0].FrameTime if ('FrameTime' in self._dsstore[0]) else 1.0

    

    @property
    def seriesInstanceUID(self) -> str:
        return self._sequenceInstanceUID

    @property
    def level_downsamples(self):
        try:
            return [self._dsstore[0].TotalPixelMatrixColumns/self._dsstore[k].TotalPixelMatrixColumns for k in self.levels]        
        except:
            return [1]

    @property 
    def level_dimensions(self):
        try:
            return [(self._dsstore[k].TotalPixelMatrixColumns,self._dsstore[k].TotalPixelMatrixRows) for k in self.levels]    
        except:
            return [self.geometry_imsize]

    def read_region_simple(self, location: tuple, level:int, size:tuple, zLevel:int=0):
        img = np.zeros((size[1],size[0],4), np.uint8)
        img[:,:,3]=255
        offset=[0,0]
        if (location[1]<0):
            offset[0] = -location[1]
            location = (location[0],0)
        if (location[0]<0):
            offset[1] = -location[0]
            location = (0,location[1])
        if (self.numberOfFrames>1) and not (self.multiInstanceStack):
            imgcut = self._dsstore[0].pixel_array[zLevel, location[1]:location[1]+size[1]-offset[0],location[0]:location[0]+size[0]-offset[1]]
        elif (self.multiInstanceStack):
            imgcut = np.copy(self._dsstore[zLevel].pixel_array[location[1]:location[1]+size[1]-offset[0],location[0]:location[0]+size[0]-offset[1]])
        else:
            imgcut = np.copy(self._dsstore[0].pixel_array[location[1]:location[1]+size[1]-offset[0],location[0]:location[0]+size[0]-offset[1]])
        imgcut = np.uint8(np.clip(np.float32(imgcut)/(self.extremes[1])*255.0,0,255))
        for k in range(3):
            img[offset[0]:imgcut.shape[0]+offset[0],offset[1]:offset[1]+imgcut.shape[1],k] = imgcut
        return Image.fromarray(img)


    @property
    def dimensions(self):
        return self.level_dimensions[0]

    def get_best_level_for_downsample(self,downsample):
        return np.argmin(np.abs(np.asarray(self.level_downsamples)-downsample))

    @property
    def level_count(self):
        return len(self.levels)

    def imagePos_to_id(self, imagePos:tuple, level:int):
        id_x, id_y = imagePos
        if (id_y>=self.geometry_rows[level]):
            id_x=self.geometry_columns[level] # out of range

        if (id_x>=self.geometry_columns[level]):
            id_y=self.geometry_rows[level] # out of range
        return (id_x+(id_y*self.geometry_columns[level]))
    
    def get_tile(self, pos, level:int):
        if pos < self._dsstore[level].NumberOfFrames:
            return np.array(Image.open(io.BytesIO(self._dsequence[level][pos])))
        else:
            if (self.channels>1):
                return np.zeros((*self.geometry_tilesize[level], self.channels))
            else:
                return np.zeros(self.geometry_tilesize[level])

    def get_id(self, pixelX:int, pixelY:int, level:int) -> (int, int, int):

        id_x = round(-0.5+(pixelX/self.geometry_tilesize[level][1]))
        id_y = round(-0.5+(pixelY/self.geometry_tilesize[level][0]))
        
        return (id_x,id_y), pixelX-(id_x*self.geometry_tilesize[level][0]), pixelY-(id_y*self.geometry_tilesize[level][1]),

        
    def read_region(self, location: tuple, level:int, size:tuple, zLevel:int=None):

        # Use simplified region reader if dicom image is able to load into memory
        if self.loadIntoMemory:
            return self.read_region_simple(location=location, level=level, size=size, zLevel=zLevel)

        # convert to overall coordinates, if not in level 0
        if (self.level_downsamples[level]>1):
            location = [int(x/self.level_downsamples[level]) for x in location]

        lu, lu_xo, lu_yo = self.get_id(*list(location),level=level)
        rl, rl_xo, rl_yo = self.get_id(*[sum(x) for x in zip(location,size)], level=level)
        # generate big image
        bigimg = 0*np.ones(((rl[1]-lu[1]+1)*self.geometry_tilesize[level][0], (rl[0]-lu[0]+1)*self.geometry_tilesize[level][1],4), np.uint8)
        bigimg[:,:,3] = 255
        for xi, xgridc in enumerate(range(lu[0],rl[0]+1)):
            for yi, ygridc in enumerate(range(lu[1],rl[1]+1)):
                if (xgridc<0) or (ygridc<0):
                    continue
                if (self.channels==3):
                    bigimg[yi*self.geometry_tilesize[level][0]:(yi+1)*self.geometry_tilesize[level][0],
                        xi*self.geometry_tilesize[level][1]:(xi+1)*self.geometry_tilesize[level][1],0:-1] = \
                        self.get_tile(self.imagePos_to_id((xgridc,ygridc),level=level), level)
                elif (self.channels==1):
                    # 
                    for k in range(3):
                        bigimg[yi*self.geometry_tilesize[level][0]:(yi+1)*self.geometry_tilesize[level][0],
                            xi*self.geometry_tilesize[level][1]:(xi+1)*self.geometry_tilesize[level][1],k] = \
                            self.get_tile(self.imagePos_to_id((xgridc,ygridc),level=level), level)
        # crop big image
        return Image.fromarray(bigimg[lu_yo:lu_yo+size[1],lu_xo:lu_xo+size[0]])




"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Annotations (definition, visualization, interaction)


"""

import cv2
import matplotlib.path as path
import numpy as np

class AnnotationType(enumerate):
    SPOT = 1
    AREA = 2
    POLYGON = 3
    SPECIAL_SPOT = 4
    CIRCLE = 5
    IMAGEANNOTATION = 6
    UNKNOWN = 255


class ViewingProfile(object):
    blindMode = False
    annotator = None
    COLORS_CLASSES = [[0,0,0,0],
                  [0,0,255,255],
                  [0,255,0,255],
                  [255,255,0,255],
                  [255,0,255,255],
                  [0,127,0,255],
                  [255,127,0,255],
                  [127,127,0,255],
                  [255,200,200,255],
                  [10, 166, 168,255],
                  [166, 10, 168,255],
                  [166,168,10,255]]
    spotCircleRadius = 25
    minimumAnnotationLabelZoom = 4
    majorityClassVote = False
    activeClasses = dict()

    def getColor(self, anno : "annotation"):
        if (anno.pluginAnnotationLabel is not None):
            return anno.pluginAnnotationLabel.color
        if (self.blindMode):
            return vp.COLORS_CLASSES[anno.labelBy(vp.annotator) ]
        elif (self.majorityClassVote):
            return self.COLORS_CLASSES[anno.majorityLabel() ]
        else:
            return self.COLORS_CLASSES[anno.agreedLabel()]


"""
  Annotation label class for plugins
"""
class PluginAnnotationLabel():

      def __set_color(self, color:tuple):
            if len(color) != 4:
                  raise ValueError('Color needs to be a tuple of four integer values.')
            elif all([isinstance(x,int) for x in color]):
                  self.__color = color
            elif all([isinstance(x,float) for x in color]):
                  self.__color = [int(x*255) for x in color]
            else:
                  raise ValueError('Color needs to be a tuple of four integer values (<255) or three float values (=<1)')
      
      def __get_color(self):
            return self.__color

      color = property(__get_color, __set_color)
                  
      def __init__(self, uid: int, name: str, color: tuple):
            self.uid = uid
            self.name = name
            self.color = color
      
      def __str__(self):
            return self.name

class AnnotationLabel(object):
    def __init__(self,annotatorId:int, classId:int, uid:int, exact_id:int=None):
        self.annnotatorId = annotatorId
        self.classId = classId
        self.uid = uid
        self.exact_id = exact_id
    

class annoCoordinate(object):
    x = None
    y = None
    def __init__(self,x,y,x_abs=None,y_abs=None):
        self.x = x
        self.y = y
        self.x_abs = x_abs
        self.y_abs = y_abs
    
    def totuple(self):
        return (self.x, self.y)
    
    def tolist(self):
        return [self.x, self.y]
    
    def __str__(self):
        return f'annoCoordinate at x={self.x}, y={self.y}'

    def __repr__(self):
        return f'annoCoordinate at x={self.x}, y={self.y}'

class AnnotationHandle(object):
    pt1 = None
    pt2 = None

    def __init__(self, pt1:annoCoordinate, pt2:annoCoordinate):
        self.pt1 = pt1
        self.pt2 = pt2

    def positionWithinRectangle(self,position:tuple, absolute:bool=False, tolerance:int=2, zoom:float=1):
        tolerance = tolerance * zoom
        if (absolute):
            return ((position[0]>=self.pt1.x_abs-tolerance) and (position[1]>=self.pt1.y_abs-tolerance) and
                    (position[0]<=self.pt2.x_abs+tolerance) and (position[1]<=self.pt2.y_abs+tolerance))
        else:
            return ((position[0]>=self.pt1.x-tolerance) and (position[1]>=self.pt1.y-tolerance) and
                    (position[0]<=self.pt2.x+tolerance) and (position[1]<=self.pt2.y+tolerance))



class annotation():

      def __init__(self, uid=0, text='',pluginAnnotationLabel=None):
          self.annotationType = AnnotationType.UNKNOWN
          self.labels = list()
          self.uid = uid
          self.text = text if text is not None else ''
          self.agreedClass = None
          self.clickable = True
          self.guid = ''
          self.deleted = 0
          self.lastModified = 0.
          self.minimumAnnotationLabelZoom = 4
          self.pluginAnnotationLabel = None
          if (pluginAnnotationLabel is not None):
              if (isinstance(pluginAnnotationLabel, PluginAnnotationLabel)):
                  self.pluginAnnotationLabel = pluginAnnotationLabel
              else:
                  raise ValueError('pluginAnnotationLabel needs to be of class PluginAnnotationLabel')

      def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int = 1, vp: ViewingProfile = ViewingProfile(), selected=False):
            return
        
      def setAgreedClass(self, agreedClass):
          self.agreedClass = agreedClass
    
      def positionInAnnotation(self, position: list, zoom:float) -> bool:
            return False

      def intersectingWithAnnotation(self, anno) -> bool:
          return self.convertToPath().intersects_path(anno.convertToPath())

      def convertToPath(self):
            return path.Path([])

      def getAnnotationsDescription(self, db) -> list:
           retval = list()
           if (self.pluginAnnotationLabel is None):
                for idx,label in enumerate(self.labels):
                    annotatorName = db.getAnnotatorByID(label.annnotatorId)
                    className = db.getClassByID(label.classId)
                    retval.append(['Anno %d' % (idx+1), '%s (%s)' % (className,annotatorName)])
                retval.append(['Agreed Class', db.getClassByID(self.agreedLabel())])
           else:
                retval.append(['Plugin class', str(self.pluginAnnotationLabel)])
                if (self.text is not None):
                    retval.append(['Description', str(self.text)])

           return retval
          
      def getBoundingBox(self) -> [int,int,int,int]:
        """
            returns the bounding box (x,y,w,h) for an object         
        """
        minC = self.minCoordinates()
        return [minC.x,minC.y] + list(self.getDimensions())

      def positionInAnnotationHandle(self, position: tuple, zoom:float) -> int:
          return None
    
          
          
      def getDescription(self, db, micronsPerPixel=None) -> list:
            if (self.pluginAnnotationLabel is None):
                return self.getAnnotationsDescription(db)
            else:
                return [['Plugin Anno',self.pluginAnnotationLabel.name],]

      def addLabel(self, label:AnnotationLabel, updateAgreed=True):
          self.labels.append(label)
          
          if (updateAgreed):
              self.agreedClass = self.majorityLabel()

      def _create_annohandle(self, image:np.ndarray, coord:tuple, markersize:int, markersize_abs:float, color:tuple, abscoord:tuple=(None,None)) -> AnnotationHandle:
            pt1_rect = (max(0,coord[0]-markersize),
                        max(0,coord[1]-markersize))
            pt2_rect = (min(image.shape[1],coord[0]+markersize),
                        min(image.shape[0],coord[1]+markersize))
            cv2.rectangle(img=image, pt1=(pt1_rect), pt2=(pt2_rect), color=[255,255,255,255], thickness=2)
            cv2.rectangle(img=image, pt1=(pt1_rect), pt2=(pt2_rect), color=color, thickness=1)

            pt1_abs = (abscoord[0]-markersize_abs, abscoord[1]-markersize_abs)
            pt2_abs = (abscoord[0]+markersize_abs, abscoord[1]+markersize_abs)
            return AnnotationHandle(annoCoordinate(*pt1_rect, *pt1_abs), annoCoordinate(*pt2_rect, *pt2_abs))

      def getDimensions(self) -> (int, int):
          minC = self.minCoordinates()
          maxC = self.maxCoordinates()
          return (int(maxC.x-minC.x),int(maxC.y-minC.y))
        
      def getCenter(self) -> (annoCoordinate):
          minC = self.minCoordinates()
          maxC = self.maxCoordinates()
          return annoCoordinate(int(0.5*(minC.x+maxC.x)),int(0.5*(minC.y+maxC.y)))
        
      def removeLabel(self, uid:int):
          for label in range(len(self.labels)):
              if (self.labels[label].uid == uid):
                  self.labels.pop(label)
     
      def changeLabel(self, uid:int, annotatorId:int, classId:int):
          for label in range(len(self.labels)):
              if (self.labels[label].uid == uid):
                  self.labels[label] = AnnotationLabel(annotatorId, classId, uid)
      
      def maxLabelClass(self):
          retVal=0
          for label in range(len(self.labels)):
              if (self.labels[label].classId > retVal):
                  retVal=self.labels[label].classId
          return retVal


      """
            Returns the majority label for an annotation
      """
      def majorityLabel(self):
          if len(self.labels)==0:
              return 0

          histo = np.zeros(self.maxLabelClass()+1)

          for label in np.arange(0, len(self.labels)):
               histo[self.labels[label].classId] += 1

          if np.sum(histo == np.max(histo))>1:
              # no clear maximum, return 0
              return 0
          else:   
              # a clear winner. Return it.
              return np.argmax(histo)


      """
            Returns the agreed (common) label for an annotation
      """
      def agreedLabel(self):
          if (self.agreedClass is not None):
              return self.agreedClass
          else:
              return 0

      
      def labelBy(self, annotatorId):
          for label in np.arange(0, len(self.labels)):
               if (self.labels[label].annnotatorId == annotatorId):
                    return self.labels[label].classId
          return 0
        
      
      def minCoordinates(self) -> annoCoordinate:
            print('Whopsy... you need to overload this.')
            return annoCoordinate(None, None)

      def maxCoordinates(self) -> annoCoordinate:
            print('Whopsy... you need to overload this.')
            return annoCoordinate(None, None)


class imageAnnotation(annotation):
      def __init__(self, uid, text='', pluginAnnotationLabel=None, minimumAnnotationLabelZoom=1):
            super().__init__(uid=uid, text=text, pluginAnnotationLabel=pluginAnnotationLabel)
            self.text = text
            self.minimumAnnotationLabelZoom = minimumAnnotationLabelZoom
            self.annotationType = AnnotationType.AREA
            self.clickable=False
    
      def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int, vp : ViewingProfile, selected = False):

            image = cv2.rectangle(image, thickness=thickness, pt1=(0,0), pt2=(image.shape[1],image.shape[0]),color=vp.getColor(self), lineType=cv2.LINE_AA)

            if (len(self.text)>0) and (zoomLevel < self.minimumAnnotationLabelZoom):
                  cv2.putText(image, self.text, (10, 10), cv2.FONT_HERSHEY_PLAIN , 0.7,vp.getColor(self),1,cv2.LINE_AA)
    
      def minCoordinates(self) -> annoCoordinate:
         return annoCoordinate(0,0)
    
      def maxCoordinates(self) -> annoCoordinate:
         return annoCoordinate(np.inf,np.inf)


class rectangularAnnotation(annotation):
      def __init__(self, uid, x1, y1, x2, y2, text='', pluginAnnotationLabel=None, minimumAnnotationLabelZoom=1):
            super().__init__(uid=uid, text=text, pluginAnnotationLabel=pluginAnnotationLabel)
            self.x1 = x1
            self.y1 = y1
            self.x2 = x2
            self.y2 = y2
            self.minimumAnnotationLabelZoom = minimumAnnotationLabelZoom
            self.annotationType = AnnotationType.AREA
            self.text = text
      
      @property
      def coordinates(self):
          return np.array([[self.x1, self.y1], [self.x2, self.y2]])

      def minCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x1, self.y1)
      
      def getDescription(self,db, micronsPerPixel=None) -> list:
          return [['Position', 'x1=%d, y1=%d, x2=%d, y2=%d' % (self.x1,self.y1,self.x2,self.y2)]] + self.getAnnotationsDescription(db)

      def maxCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x2, self.y2)

      def positionInAnnotation(self, position: list, zoom:float) -> bool:
            return ((position[0]>self.x1) and (position[0]<self.x2) and 
                   (position[1]>self.y1) and (position[1]<self.y2))

      def width(self) -> float:
          return self.x2-self.x1

      def height(self) -> float:
          return self.y2-self.y1
        
      def centerCoordinate(self) -> tuple:
          return annoCoordinate(x=int(0.5*(self.x1+self.x2)),y=int(0.5*(self.y1+self.y2)))

      def convertToPath(self):
          return path.Path(np.array([[self.x1, self.x2, self.x2, self.x1, self.x1] ,[self.y1, self.y1, self.y2, self.y2, self.y1]]).T)

      def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int, vp : ViewingProfile = None, selected = False):
            xpos1=max(0,int((self.x1-leftUpper[0])/zoomLevel))
            ypos1=max(0,int((self.y1-leftUpper[1])/zoomLevel))
            xpos2=min(image.shape[1],int((self.x2-leftUpper[0])/zoomLevel))
            ypos2=min(image.shape[0],int((self.y2-leftUpper[1])/zoomLevel))

            image = cv2.rectangle(image, thickness=thickness, pt1=(xpos1,ypos1), pt2=(xpos2,ypos2),color=vp.getColor(self), lineType=cv2.LINE_AA)

            if self.text is not None and (len(self.text)>0) and (zoomLevel < self.minimumAnnotationLabelZoom):
                  cv2.putText(image, self.text, (xpos1+3, ypos2+10), cv2.FONT_HERSHEY_PLAIN , 0.7,(0,0,0),1,cv2.LINE_AA)

class polygonAnnotation(annotation):
    def __init__(self, uid:int, coordinates: np.ndarray = None, text='', pluginAnnotationLabel=None, minimumAnnotationLabelZoom=1):
        super().__init__(uid=uid, pluginAnnotationLabel=pluginAnnotationLabel, text=text)
        self.annotationType = AnnotationType.POLYGON
        self.annoHandles = list()
        self.minimumAnnotationLabelZoom = minimumAnnotationLabelZoom
        if (coordinates is not None):
            self.coordinates = coordinates
    
    def positionInAnnotationHandle(self, position: tuple, zoom:float) -> int:
        for key,annoHandle in enumerate(self.annoHandles):
             if (annoHandle.positionWithinRectangle(position,absolute=True,zoom=zoom)):
                 return key
        return None

    def minCoordinates(self) -> annoCoordinate:
        return annoCoordinate(self.coordinates[:,0].min(), self.coordinates[:,1].min())

    def maxCoordinates(self) -> annoCoordinate:
        return annoCoordinate(self.coordinates[:,0].max(), self.coordinates[:,1].max())
    
    def area_px(self) -> float:
        print(np.array(self.coordinates).shape,np.array(self.coordinates).dtype)
        return cv2.contourArea(np.int32(np.array(self.coordinates)))

    # Largest diameter --> 2* radius of minimum enclosing circle
    def diameter_px(self) -> float:
        (x,y),radius = cv2.minEnclosingCircle(np.int32(np.array(self.coordinates)))
        return radius*2

    def getDescription(self,db, micronsPerPixel=None) -> list:
        mc = annoCoordinate(self.coordinates[:,0].mean(), self.coordinates[:,1].mean())
        area_px = float(self.area_px())
        diameter_px = float(self.diameter_px())
        micronsPerPixel = float(micronsPerPixel)
        if micronsPerPixel < 2E-6:
            area = '%d px^2' % area_px
            diameter = '%d px' % diameter_px
        else:
            area_mum2 = (area_px*micronsPerPixel*micronsPerPixel)
            diameter_mum = diameter_px*micronsPerPixel
            if (area_mum2 < 1E4):
                area = '%.2f µm^2' % area_mum2
            else:
                area = '%.2f mm^2' % (1E-6 * area_mum2)
            if (diameter_mum < 1e3):
                diameter = '%.2f µm' % diameter_mum
            else:
                diameter = '%.2f mm' % (diameter_mum * 1e-3)

        return [['Position', 'x1=%d, y1=%d' % (mc.x,mc.y)], ['Area', area], ['Largest diameter', diameter]] + self.getAnnotationsDescription(db)

    def convertToPath(self):
        p = path.Path(self.coordinates)
        return p

    def positionInAnnotation(self, position: list, zoom:float) -> bool:
        return self.convertToPath().contains_point(position) or self.positionInAnnotationHandle(position, zoom=zoom) is not None

    def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int, vp : ViewingProfile, selected=False):
        def slideToScreen(pos):
            """
                convert slide coordinates to screen coordinates
            """
            xpos,ypos = pos
            p1 = leftUpper
            cx = int((xpos - p1[0]) / zoomLevel)
            cy = int((ypos - p1[1]) / zoomLevel)
            return (cx,cy)        
        markersize = 4#min(4,int(5/zoomLevel))
        listIdx=-1

        self.annoHandles=list()

        # small assertion to fix bug #12
        if (self.coordinates.shape[1]==0):
            return image

        for listIdx in range(self.coordinates.shape[0]-1):
            anno = slideToScreen(self.coordinates[listIdx])
            cv2.line(img=image, pt1=anno, pt2=slideToScreen(self.coordinates[listIdx+1]), thickness=2, color=vp.getColor(self), lineType=cv2.LINE_AA)       

            if (selected):
                self.annoHandles.append(self._create_annohandle(image, anno, markersize, markersize*zoomLevel, vp.getColor(self), abscoord=self.coordinates[listIdx]))


        listIdx+=1
        anno = slideToScreen(self.coordinates[listIdx])
        if (selected):
                self.annoHandles.append(self._create_annohandle(image, anno, markersize, markersize*zoomLevel,  vp.getColor(self), abscoord=self.coordinates[listIdx]))

        cv2.line(img=image, pt1=anno, pt2=slideToScreen(self.coordinates[0]), thickness=2, color=vp.getColor(self), lineType=cv2.LINE_AA)       

        if (len(self.text)>0) and (zoomLevel < self.minimumAnnotationLabelZoom):
                xpos1=int(0.5*(np.max(self.coordinates[:,0])+np.min(self.coordinates[:,0]) ))
                ypos1=int(0.5*(np.max(self.coordinates[:,1])+np.min(self.coordinates[:,1])))
                cv2.putText(image, self.text, slideToScreen((xpos1+3, ypos1+10)), cv2.FONT_HERSHEY_PLAIN , 0.7,(0,0,0),1,cv2.LINE_AA)



class circleAnnotation(annotation):
      
      def __init__(self, uid, x1, y1, x2 = None, y2 = None, r = None, text='', pluginAnnotationLabel=None, minimumAnnotationLabelZoom=1):
            super().__init__(uid=uid, text=text, pluginAnnotationLabel=pluginAnnotationLabel)
            self.annotationType = AnnotationType.CIRCLE
            self.minimumAnnotationLabelZoom = minimumAnnotationLabelZoom

            if (r is None):
                self.x1 = int(0.5*(x1+x2))
                self.y1 = int(0.5*(y1+y2))
                if (x2>x1):
                    self.r = int((x2-x1)*0.5)
                else:
                    self.r = int((x1-x2)*0.5)
            else:
                self.x1 = int(x1)
                self.y1 = int(y1)
                self.r = int(r)

      @property
      def coordinates(self):
          return np.array([[self.x1-self.r, self.y1-self.r], [self.x1+self.r, self.y1+self.r]])

      def minCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x1-self.r, self.y1-self.r)

      def convertToPath(self):
          pi = np.linspace(0,2*np.pi,100)
          x = np.sin(pi)*self.r+self.x1
          y = np.cos(pi)*self.r+self.y1
          return path.Path(np.c_[x,y])
          

      def maxCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x1+self.r, self.y1+self.r)

      def getDescription(self,db, micronsPerPixel=None) -> list:
          return [['Position', 'x1=%d, y1=%d' % (self.x1, self.y1)]] + self.getAnnotationsDescription(db)

      def positionInAnnotation(self, position: list,zoom:float) -> bool:
          dist = np.sqrt(np.square(position[0]-self.x1)+np.square(position[1]-self.y1))
          return (dist<=self.r)

      def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int, vp : ViewingProfile, selected=False):
            xpos1=int((self.x1-leftUpper[0])/zoomLevel)
            ypos1=int((self.y1-leftUpper[1])/zoomLevel)
            radius = int(self.r/zoomLevel)
            if (radius>=0):
                  image = cv2.circle(image, thickness=thickness, center=(xpos1,ypos1), radius=radius,color=vp.getColor(self), lineType=cv2.LINE_AA)
            if (len(self.text)>0) and (zoomLevel < self.minimumAnnotationLabelZoom):
                    cv2.putText(image, self.text, (xpos1,ypos1), cv2.FONT_HERSHEY_PLAIN , 0.7,(0,0,0),1,cv2.LINE_AA)


class spotAnnotation(annotation):

      def __init__(self, uid, x1, y1, isSpecialSpot : bool = False,text='', pluginAnnotationLabel=None, minimumAnnotationLabelZoom=1, radius=None):
            super().__init__(uid=uid, text=text, pluginAnnotationLabel=pluginAnnotationLabel)
            self.annotationType = AnnotationType.SPOT
            self.x1 = x1
            self.y1 = y1
            self.r = 25 if radius is None else radius
            if (isSpecialSpot):
                self.annotationType = AnnotationType.SPECIAL_SPOT
            self.minimumAnnotationLabelZoom = minimumAnnotationLabelZoom

      def minCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x1-25, self.y1-25)

      def maxCoordinates(self) -> annoCoordinate:
            return annoCoordinate(self.x1+25, self.y1+25)

      def intersectingWithAnnotation(self, anno) -> bool:
            return False

      @property
      def coordinates(self):
          return np.array([[self.x1-self.r, self.y1-self.r], [self.x1+self.r, self.y1+self.r]])


      def draw(self, image: np.ndarray, leftUpper: tuple, zoomLevel: float, thickness: int, vp : ViewingProfile, selected=False):
            xpos1=int((self.x1-leftUpper[0])/zoomLevel)
            ypos1=int((self.y1-leftUpper[1])/zoomLevel)
            radius=int(int(vp.spotCircleRadius)/zoomLevel)
            if (radius>=0):
                  image = cv2.circle(image, thickness=thickness, center=(xpos1,ypos1), radius=radius,color=vp.getColor(self), lineType=cv2.LINE_AA)
            if (len(self.text)>0) and (zoomLevel < self.minimumAnnotationLabelZoom):
                    cv2.putText(image, self.text, (xpos1+3, ypos1+10), cv2.FONT_HERSHEY_PLAIN , 0.7,(0,0,0),1,cv2.LINE_AA)

      def getDescription(self,db, micronsPerPixel=None) -> list:
          return [['Position', 'x1=%d, y1=%d' % (self.x1, self.y1)]] + self.getAnnotationsDescription(db)

      def positionInAnnotation(self, position: list, zoom:float) -> bool:
          dist = np.sqrt(np.square(position[0]-self.x1)+np.square(position[1]-self.y1))
          return (dist<=25)



"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, 
         Technische Hochschule Ingolstadt
         marc.aubreville@thi.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images. 
        In: Bildverarbeitung für die Medizin 2018. 
        Springer Vieweg, Berlin, Heidelberg, 2018. pp. 309-314.


   This file: Support for NIFTI files. Requires nibabel to work.


"""

import numpy as np
try:
    import nibabel as nib 
    nii_fileformats = ['*.nii', '*.nii.gz']
except:
    nib = None
    print('Warning: nibabel not installed. Neuroimaging formats will not be available.')
    nii_fileformats = []

from PIL import Image
import io
import os



class ReadableNIBDataset():
    def __init__(self, filename):

        self._sequenceInstanceUID = ''
        
        self._path, _ = os.path.split(filename)
        if (self._path==""):
            self._path='.'+os.sep

        self._dsstore = dict()

        self.isOpenSlide = False
        self.nib = nib.load(filename)
        self.imdata = self.nib.dataobj.reshape((*self.nib.shape[0:2],-1)) # flatten 
        
        self.levels = [0]
        self.channels = 1
        self.numberOfFrames = self.imdata.shape[2] # last dimension is number of frames

        self.mpp_x = 1000/self.nib.header.get_zooms()[1] # 1000 microns/mm / (vx/mm) = microns/vx 
        self.geometry_imsize = self.imdata.shape[0:2]
        self.geometry_rows = [1]
        self.geometry_columns = [1]
        self.extremes = np.min(self.imdata), np.max(self.imdata)

    

    @property
    def level_downsamples(self):
        return [1]

    @property 
    def level_dimensions(self):
        return [self.geometry_imsize]

    def read_region(self, location: tuple, level:int, size:tuple, zLevel:int=0, rotate:bool=True):
        img = np.zeros((size[1],size[0],4), np.uint8)
        img[:,:,3]=255
        offset=[0,0]
        if (location[1]<0):
            offset[0] = -location[1]
            location = (location[0],0)
        if (location[0]<0):
            offset[1] = -location[0]
            location = (0,location[1])
        imgcut = self.imdata[location[1]:location[1]+size[1]-offset[0],location[0]:location[0]+size[0]-offset[1],zLevel]
        ext = np.percentile(self.imdata[:,:,zLevel],5),np.percentile(self.imdata[:,:,zLevel],95)
#        print('Extremes:',ext,'rotate:',rotate,'zLevel',zLevel)
        imgcut = np.uint8(np.clip(np.float32(imgcut)/(ext[1])*255.0,0,255))
        for k in range(3):
            img[offset[0]:imgcut.shape[0]+offset[0],offset[1]:offset[1]+imgcut.shape[1],k] = imgcut
        return Image.fromarray(img)


    @property
    def dimensions(self):
        return self.level_dimensions[0]

    def get_best_level_for_downsample(self,downsample):
        return np.argmin(np.abs(np.asarray(self.level_downsamples)-downsample))

    @property
    def level_count(self):
        return len(self.levels)

    def imagePos_to_id(self, imagePos:tuple, level:int):
        id_x, id_y = imagePos
        if (id_y>=self.geometry_rows[level]):
            id_x=self.geometry_columns[level] # out of range

        if (id_x>=self.geometry_columns[level]):
            id_y=self.geometry_rows[level] # out of range
        return (id_x+(id_y*self.geometry_columns[level]))
    

        
    